from biolib.typing_utils import List, Optional

class SeqUtilRecord:
    def __init__(self, sequence: str, sequence_id: str, description: Optional['str']):
        self.sequence = sequence
        self.id = sequence_id # pylint: disable=invalid-name
        self.description = description

class SeqUtil:
    @staticmethod
    def parse_fasta(file_name: str, default_header: Optional['str'] = None) -> List['SeqUtilRecord']:
        with open(file_name, 'r') as file_handle:
            data = file_handle.read().strip()
        if not data:
            return []
        if '>' not in data:
            if default_header:
                data = f'>{default_header}\n{data}'
            else:
                raise Exception('No header line found in fasta file')
        splitted = []
        tmp_data = ""
        for line in data.splitlines():
            if line.startswith(">"):
                if tmp_data:
                    splitted.append(tmp_data)
                tmp_data = line[1:].strip() + '\n'
            else:
                if line.strip():
                    tmp_data += line.strip() + '\n'
        if tmp_data:
            splitted.append(tmp_data)
        parsed_sequences = []
        for sequence_data in splitted:
            sequence_data_splitted = sequence_data.strip().split('\n')
            header_line = sequence_data_splitted[0].split()
            sequence_id = header_line[0]
            description = sequence_data_splitted[0][len(sequence_id):].strip()
            sequence = "".join([seq.strip() for seq in sequence_data_splitted[1:]])
            parsed_sequences.append(
                SeqUtilRecord(sequence=sequence, sequence_id=sequence_id, description=description)
            )
        return parsed_sequences



from biolib import utils
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_binary_format import UnencryptedModuleOutput
from biolib.biolib_logging import logger
from biolib.compute_node.job_worker.executors.types import RemoteExecuteOptions

from biolib.compute_node.job_worker.job_storage import JobStorage
from biolib.jobs import Job


class RemoteExecutor:

    @staticmethod
    def execute_job(options: RemoteExecuteOptions, module_input_serialized: bytes) -> UnencryptedModuleOutput:
        job_id = options['job']['public_id']
        JobStorage.upload_module_input(
            job=options['job'],
            module_input_serialized=module_input_serialized
        )
        cloud_job = BiolibJobApi.create_cloud_job(job_id=job_id, result_name_prefix=options['result_name_prefix'])
        logger.debug(f"Cloud: Job created with id {cloud_job['public_id']}")

        if utils.IS_RUNNING_IN_NOTEBOOK:
            utils.STREAM_STDOUT = True

        app_version = options['job']['app_version']
        enable_print = bool(
            utils.STREAM_STDOUT and
            (app_version.get('main_output_file') or app_version.get('stdout_render_type') == 'text')
        )
        job = Job.create_from_uuid(
            uuid=options['job']['public_id'],
            auth_token=options['job']['auth_token']
        )
        job._stream_logs(enable_print)  # pylint: disable=protected-access

        BiolibApiClient.refresh_auth_token()
        return JobStorage.get_module_output(job=options['job'])



import tempfile

from biolib import utils
from biolib.biolib_binary_format import UnencryptedModuleOutput, InMemoryIndexableBuffer
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.compute_node.job_worker.executors import RemoteExecutor, RemoteExecuteOptions
from biolib.compute_node.job_worker.job_worker import JobWorker
from biolib.typing_utils import Optional


def run_job(
        job,
        module_input_serialized: bytes,
        force_local: bool = False,
        result_name_prefix: Optional[str] = None,
) -> UnencryptedModuleOutput:
    job_id = job['public_id']
    logger.info(f'Job "{job_id}" is starting...')

    if not force_local:
        return RemoteExecutor.execute_job(
            RemoteExecuteOptions(
                biolib_base_url=utils.BIOLIB_BASE_URL,
                job=job,
                result_name_prefix=result_name_prefix,
                root_job_id=job_id,
            ),
            module_input_serialized,
        )

    if not job['app_version'].get('modules'):
        BioLibError('Unable to run the application locally')

    with tempfile.TemporaryDirectory() as job_temporary_dir:
        module_output_path = JobWorker().run_job_locally(job, module_input_serialized, job_temporary_dir)

        with open(module_output_path, mode='rb') as module_output_file:
            module_output_serialized = module_output_file.read()

        return UnencryptedModuleOutput(InMemoryIndexableBuffer(module_output_serialized))



from biolib.biolib_binary_format import BioLibBinaryFormatBasePackage
from biolib.biolib_binary_format.utils import IndexableBuffer, InMemoryIndexableBuffer, LazyLoadedFile
from biolib.biolib_binary_format.file_in_container import FileInContainer
from biolib.typing_utils import TypedDict, List, Optional


class Metadata(TypedDict):
    version: int
    type: int
    stdout_length: int
    stderr_length: int
    files_info_length: int
    files_data_length: int
    exit_code: int


class UnencryptedModuleOutput(BioLibBinaryFormatBasePackage):
    _version = 1
    _type = 11
    _metadata_byte_lengths = dict(
        # Note: order is important
        version=1,
        type=1,
        stdout_length=8,
        stderr_length=8,
        files_info_length=8,
        files_data_length=8,
        exit_code=2,
    )
    _metadata_length = sum(_metadata_byte_lengths.values())
    _file_path_length_bytes = 4
    _file_data_length_bytes = 8

    def __init__(self, buffer: IndexableBuffer):
        super().__init__()
        self._buffer = buffer

        self._metadata: Optional[Metadata] = None
        self._stdout: Optional[bytes] = None
        self._stderr: Optional[bytes] = None
        self._files: Optional[List[LazyLoadedFile]] = None

    @property
    def buffer(self) -> IndexableBuffer:
        return self._buffer

    def get_exit_code(self) -> int:
        metadata = self._get_metadata()
        return metadata['exit_code']

    def get_stdout(self) -> bytes:
        if self._stdout is None:
            metadata = self._get_metadata()
            self._stdout = self._buffer.get_data(start=self._metadata_length, length=metadata['stdout_length'])

        return self._stdout

    def get_stderr(self) -> bytes:
        if self._stderr is None:
            metadata = self._get_metadata()
            self._stderr = self._buffer.get_data(
                start=self._metadata_length + metadata['stdout_length'],
                length=metadata['stderr_length'],
            )

        return self._stderr

    def get_files(self) -> List[LazyLoadedFile]:
        metadata = self._get_metadata()
        if self._files is None:
            self._files = []
            if metadata['files_info_length'] == 0:
                return self._files

            files_info_start = self._metadata_length + metadata['stdout_length'] + metadata['stderr_length']
            files_info_buffer = InMemoryIndexableBuffer(
                data=self._buffer.get_data(start=files_info_start, length=metadata['files_info_length'])
            )

            files_data_pointer = files_info_start + metadata['files_info_length']
            while files_info_buffer.pointer < len(files_info_buffer):
                path_length = files_info_buffer.get_data_with_pointer_as_int(self._file_path_length_bytes)
                path = files_info_buffer.get_data_with_pointer_as_string(path_length)
                data_length = files_info_buffer.get_data_with_pointer_as_int(self._file_data_length_bytes)

                data_start = files_data_pointer
                files_data_pointer += data_length

                self._files.append(LazyLoadedFile(path=path, buffer=self._buffer, start=data_start, length=data_length))

        return self._files

    def _get_metadata(self) -> Metadata:
        if self._metadata is None:
            metadata_buffer = InMemoryIndexableBuffer(self._buffer.get_data(start=0, length=self._metadata_length))

            partial_metadata = {}
            for field_name, field_length in self._metadata_byte_lengths.items():
                value = metadata_buffer.get_data_with_pointer_as_int(length=field_length)  # type: ignore
                if field_name == 'version' and value != UnencryptedModuleOutput._version:
                    raise Exception('Version does not match')

                if field_name == 'type' and value != UnencryptedModuleOutput._type:
                    raise Exception('Type does not match')

                partial_metadata[field_name] = value

            self._metadata = partial_metadata  # type: ignore

        return self._metadata  # type: ignore

    @staticmethod
    def write_to_file(
            output_file_path: str,
            exit_code: int,
            files: List[FileInContainer],
            stderr: bytes,
            stdout: bytes,
    ) -> None:
        with open(output_file_path, mode='wb') as output_file:
            meta_lengths = UnencryptedModuleOutput._metadata_byte_lengths

            output_file.write(UnencryptedModuleOutput._version.to_bytes(meta_lengths['version'], byteorder='big'))
            output_file.write(UnencryptedModuleOutput._type.to_bytes(meta_lengths['type'], byteorder='big'))
            output_file.write(len(stdout).to_bytes(meta_lengths['stdout_length'], byteorder='big'))
            output_file.write(len(stderr).to_bytes(meta_lengths['stdout_length'], byteorder='big'))

            files_info = bytearray()
            files_data_size = 0

            for file in files:
                file_size = file.get_data_size_in_bytes()
                path_as_bytes = file.path.encode()
                path_length_as_bytes = len(path_as_bytes).to_bytes(length=4, byteorder='big')
                data_byte_length_as_bytes = file_size.to_bytes(length=8, byteorder='big')

                files_info.extend(path_length_as_bytes + path_as_bytes + data_byte_length_as_bytes)
                files_data_size += file_size

            output_file.write(len(files_info).to_bytes(length=8, byteorder='big'))
            output_file.write(files_data_size.to_bytes(length=8, byteorder='big'))
            output_file.write(exit_code.to_bytes(length=meta_lengths['exit_code'], byteorder='big'))
            output_file.write(stdout)
            output_file.write(stderr)
            output_file.write(files_info)

            for file in files:
                for chunk in file.get_data_stream():
                    output_file.write(chunk)
